from __future__ import absolute_import

def get_package_data():
    return {'gala.coordinates': ['tests/idl_vgsr_vhel.txt',
                                 'tests/SgrCoord_data',
                                 'tests/gd1_coord.txt']}
